import se.sics.isl.transport.Transportable;
import se.sics.tasim.logtool.LogReader;
import se.sics.tasim.logtool.ParticipantInfo;
import se.sics.tasim.props.*;
import se.sics.tasim.tac05.Parser;
import java.util.*;
import java.io.*;

/**
 * Parses a single game log.  All incoming and outgoing messages for the 
 * specified agent are stored, along with all prices for which predictions
 * are needed.  The predictor is then called with each day's information, and
 * the predictions are evaluated.
 */
public class PredictionServerParser extends Parser {

	final static int N_COMPUTER = 20; //how many days in the future to predict
	final static int N_COMPONENT = 20; //how many days in the future to predict
		
	final static int numDays = 220;
	final static int numComputers = 16;
	final static int numComponents = 10;
	
	private String agentName; //name of the agent to track
	private boolean verbose = false;
	private Vector[] messagesIn = new Vector[numDays];
	private Vector[] messagesOut = new Vector[numDays];
	private Vector[] futureRFQs = new Vector[numDays];
	
	private ServerSocketHandler sh;
	
	int date = 0;
	ParticipantInfo[] participants;
	
	BOMBundle bom;
	ComponentCatalog catalog;			
	int[] computerIDs = new int[numComputers];
	int[] componentIDs = new int[numComponents];
	int[][] componentIndexList = new int[numComputers][4]; //4 components per computer
	int[] computerBasePrices = new int[numComputers];
	int[] componentBasePrices = new int[numComponents];
	
	double[][] medianPrice = new double[numComputers][numDays];
	Vector[] todaysPrices = new Vector[numComputers];
	
	Vector[] supplierOffers = new Vector[numDays];
	Hashtable supplierOrders = new Hashtable();
	Hashtable supplierRFQs = new Hashtable();
	Hashtable customerOffers = new Hashtable();
	Vector[] customerOrders = new Vector[numDays];
	Hashtable customerRFQs = new Hashtable();
	
	//counts of predictions and sums of errors
	int futureComputerPredictionCount = 0;
	double futureComputerPredictionSquaredError = 0;
	int computerPredictionCount = 0;
	double computerPredictionSquaredError = 0;
	int futureComponentPredictionCount = 0;
	double futureComponentPredictionSquaredError = 0;
	int componentPredictionCount = 0;
	double componentPredictionSquaredError = 0;
	
	double[] results;
	PrintStream output;

	/**
	 * Called by the main class for each game.
	 * 
	 * @param reader  reads the game log
	 * @param agentName  the agent for which predictions are made
	 * @param predictor  makes predictions	
	 * @param verbose  whether to output all predictions
	 * @param results  accuracy results to be returned
	 */
	public PredictionServerParser(LogReader reader, String agentName, ServerSocketHandler socketHandler, PrintStream output, boolean verbose, double[] results) {
		super(reader);
		this.agentName = agentName;
		sh = socketHandler;
		this.verbose = verbose;
		this.results = results;
		this.output = output;
		
		participants = reader.getParticipants();
		if (participants == null) {
			throw new IllegalStateException("no participants");
		}
		
		for (int d = 0; d < numDays; d++){	
			messagesIn[d] = new Vector();
			messagesOut[d] = new Vector();
			futureRFQs[d] = new Vector();
			customerOrders[d] = new Vector();
			supplierOffers[d] = new Vector();
		}
	}
	
	
	/**
	 * Called each time a message is read from the game log.
	 */
	protected void message(int sender, int receiver, Transportable content) {
		if (date >= numDays)//ignore any messages at the end
			return;
		
		//first store messages to the agent
		if (participants[receiver].getName().equals(agentName)){
			Message m = new Message(participants[sender].getName(), participants[receiver].getName(), content);
			messagesIn[date].add(m);
		}
		else if (participants[sender].getName().equals(agentName)){
			Message m = new Message(participants[sender].getName(), participants[receiver].getName(), content);
			messagesOut[date].add(m);
		}
		
		//now store information needed to check predictions
		if (participants[sender].getName().equals(agentName) && content instanceof RFQBundle){
			//this is an RFQ bundle for a supplier from the agent
			handleSRFQ(sender, receiver, (RFQBundle) content);
		}
		else if (participants[receiver].getRole() == MANUFACTURER && content instanceof OrderBundle){
			//this is an order bundle for a manufacturer from customers
			handleCOrder(sender, receiver, (OrderBundle) content);	
		}
		else if (participants[receiver].getName().equals(agentName) && content instanceof OfferBundle){
			//this is an offer bundle for a manufacturer from a supplier	
			handleSOffer(sender, receiver, (OfferBundle) content);
		}
		else if (participants[receiver].getRole() == CUSTOMER && content instanceof OfferBundle){
			//this is an offer bundle for customers from a manufacturer
			handleCOffer(sender, receiver, (OfferBundle) content);	
		}
		else if (content instanceof RFQBundle && participants[receiver].getName().equals(agentName)){
			//this is an rfq bundle from customers
			handleCRFQ(sender, receiver, (RFQBundle) content);	
		}
		
		
	}

	
	/**
	 * BOMBundle, ComponentCatalog, and StartInfo are given here, instead of
	 * with the messages
	 */
	protected void dataUpdated(int type, Transportable content) {
		if (content instanceof BOMBundle){
			bom = (BOMBundle) content;
		}
		else if (content instanceof ComponentCatalog){
			catalog = (ComponentCatalog) content;
		}
		else if (content instanceof StartInfo){
			//this is StartInfo
		}
	}

	/**
	 * Called each time a new day is reached in the game log
	 */
	protected void nextDay(int somedate, long serverTime) {
		date = somedate;
			
		if (date == 0){ 
			//for some reason the parser handles the BOM and catalog differently from messages,
			//so create messages for them on the first day
			Message m = new Message("server", agentName, bom);
			messagesIn[date].add(m);
			m = new Message("server", agentName, catalog);
			messagesIn[date].add(m);
			
			//now do any necessary setup using the BOM and catalog
			setup();
		}
		
		if (date >= 1)
			getMedianPrices();
			
	}
	
	/**
	 * Called at the end of the game.  
	 */
	protected void parseStopped() {
		//the game has been parsed, so now get the predictions
		getPredictions();
	}

	
	/**
	 * After the game is parsed, this will get the daily predictions from the
	 * predictor. 
	 */
	private void getPredictions(){
		for (int d = 0; d < numDays; d++){
			PredictionObject predictions = new PredictionObject();
			
			//only need to predict future computer prices for up to N_COMPUTER+2 days before the end
			boolean futureComputerPriceNeeded = d < numDays - N_COMPUTER - 2;
			predictions.setFutureComputerPriceNeeded(futureComputerPriceNeeded);
			
			//no computer rfq price predictions needed on the first or last two days
			boolean computerPriceNeeded = d > 0 && d < numDays - 2;
			predictions.setComputerPriceNeeded(computerPriceNeeded);
			
			//add list of future RFQs for which predictions are needed
			Message[] mlist;
			if (d+N_COMPONENT < numDays - 10){ //no RFQs last 10 days
				mlist = new Message[futureRFQs[d+N_COMPONENT].size()];
				futureRFQs[d+N_COMPONENT].toArray(mlist);
			}
			else
				mlist = new Message[0];
			predictions.setFutureComponentRFQList(mlist);
			
			//get the predictions
			predictions = (PredictionObject) sh.writeAndRead(new DailyInfo(d, messagesIn[d], messagesOut[d], predictions));
			
			//evaluate predictions
			int dailyFutureComputerPredictionCount = 0;
			double dailyFutureComputerPredictionSquaredError = 0;
			int dailyComputerPredictionCount = 0;
			double dailyComputerPredictionSquaredError = 0;
			int dailyFutureComponentPredictionCount = 0;
			double dailyFutureComponentPredictionSquaredError = 0;
			int dailyComponentPredictionCount = 0;
			double dailyComponentPredictionSquaredError = 0;
						
			if (verbose)
				output.println("========================\nDAY " + d + "\n========================");
			
			//check current computer price predictions
			if (computerPriceNeeded){
				if (verbose)
					output.println("CURRENT COMPUTER PRICES:");
			
				for (int i = 0; i < customerOrders[d].size(); i++){
					OrderInfo order = (OrderInfo) customerOrders[d].get(i);
					double predicted = predictions.getComputerRFQPrice(order.rfq.rfqID) / computerBasePrices[order.index];
					if (Double.isNaN(predicted)) //no prediction, so just use 0
						predicted = 0;
					double actual = order.price / (double) computerBasePrices[order.index];
					//if (verbose)
					//	output.println("RFQ: " + order.rfq.rfqID + "\t actual price: " + pad(actual) + "\t prediction: " + pad(predicted));
					double errorSquared = Math.pow(actual - predicted, 2);
					dailyComputerPredictionSquaredError += errorSquared;
					computerPredictionSquaredError += errorSquared;
					dailyComputerPredictionCount++;
					computerPredictionCount++;
				}
				
				if (verbose && dailyComputerPredictionCount > 0)
					output.println("DAILY RMS ERROR: " + Math.sqrt(dailyComputerPredictionSquaredError/dailyComputerPredictionCount) + "\n");
				
			}
			
			//check future computer price predictions
			if (futureComputerPriceNeeded){
				if (verbose)
					output.println("FUTURE COMPUTER PRICES:");
				for (int i = 0; i < numComputers; i++){
					if (!Double.isNaN(medianPrice[i][d+N_COMPUTER])){ //median exists
						int pid = computerIDs[i];
						double actual = medianPrice[i][d+N_COMPUTER];
						double predicted = predictions.getFutureMedianComputerPrice(pid) / computerBasePrices[i]; 
						//if (verbose)
						//	output.println("ID: " + pid + "\t actual median: " + pad(actual) + "\t prediction: " + pad(predicted));
						double errorSquared = Math.pow(actual - predicted, 2);
						dailyFutureComputerPredictionSquaredError += errorSquared;
						futureComputerPredictionSquaredError += errorSquared;
						dailyFutureComputerPredictionCount++;
						futureComputerPredictionCount++;
					}
				}
				if (verbose && dailyFutureComputerPredictionCount > 0)
					output.println("DAILY RMS ERROR: " + Math.sqrt(dailyFutureComputerPredictionSquaredError/dailyFutureComputerPredictionCount) + "\n");
			}
			
			//check current component price predictions
			if (supplierOffers[d].size() > 0){
				if (verbose)
					output.println("CURRENT COMPONENT PRICES:");
			
				for (int i = 0; i < supplierOffers[d].size(); i++){
					OfferInfo offer = (OfferInfo) supplierOffers[d].get(i);
					//ignore if due date or quantity was changed
					if (offer.dueDate == offer.rfq.dueDate && offer.quantity == offer.rfq.quantity){
						double predicted = predictions.getComponentRFQPrice(offer.rfq.rfqID) / componentBasePrices[offer.index];
						if (Double.isNaN(predicted)) //no prediction, so just use 0
							predicted = 0;
						double actual = offer.price / (double) componentBasePrices[offer.index];
						//if (verbose)
						//	output.println("RFQ: " + offer.rfq.rfqID + "\t actual price: " + pad(actual) + "\t prediction: " + pad(predicted));
						double errorSquared = Math.pow(actual - predicted, 2);
						dailyComponentPredictionSquaredError += errorSquared;
						componentPredictionSquaredError += errorSquared;
						dailyComponentPredictionCount++;
						componentPredictionCount++;
					}
				}
				
				if (verbose && dailyComponentPredictionCount > 0)
					output.println("DAILY RMS ERROR: " + Math.sqrt(dailyComponentPredictionSquaredError/dailyComponentPredictionCount) + "\n");
			}
			
			//check future component price predictions
			if (d + N_COMPONENT < numDays - 10 && supplierOffers[d + N_COMPONENT].size() > 0){
				if (verbose)
					output.println("FUTURE COMPONENT PRICES:");
			
				for (int i = 0; i < supplierOffers[d + N_COMPONENT].size(); i++){
					OfferInfo offer = (OfferInfo) supplierOffers[d + N_COMPONENT].get(i);
					if (offer.rfq.rfqID < 0){ //rfqs for which future price predicitons are needed have ids < 0
						double predicted = predictions.getFutureComponentRFQPrice(offer.rfq.rfqID) / componentBasePrices[offer.index];
						if (Double.isNaN(predicted)) //no prediction, so just use 0
							predicted = 0;
						double actual = offer.price / (double) componentBasePrices[offer.index];
						//if (verbose)
						//	output.println("RFQ: " + offer.rfq.rfqID + "\t actual price: " + pad(actual) + "\t prediction: " + pad(predicted));
						double errorSquared = Math.pow(actual - predicted, 2);
						dailyFutureComponentPredictionSquaredError += errorSquared;
						futureComponentPredictionSquaredError += errorSquared;
						dailyFutureComponentPredictionCount++;
						futureComponentPredictionCount++;
					}
				}
				
				if (verbose && dailyFutureComponentPredictionCount > 0)
					output.println("DAILY RMS ERROR: " + Math.sqrt(dailyFutureComponentPredictionSquaredError/dailyFutureComponentPredictionCount) + "\n");
			}
		}
		
		//now get average errors for the game
		//results array: current computer, future computer, current component, future component RMS error
		results[0] = Math.sqrt(computerPredictionSquaredError/computerPredictionCount);
		results[1] = Math.sqrt(futureComputerPredictionSquaredError/futureComputerPredictionCount);
		results[2] = Math.sqrt(componentPredictionSquaredError/componentPredictionCount);
		results[3] = Math.sqrt(futureComponentPredictionSquaredError/futureComponentPredictionCount);
	}
	
	/**
	 * Sets up information using the catalog and BOM.
	 */
	private void setup(){
		for (int i = 0; i < numComponents; i++)
			componentIDs[i] = catalog.getProductID(i);
		for (int i = 0; i < numComputers; i++)
			computerIDs[i] = bom.getProductID(i);

		for (int i = 0; i < numComponents; i++){
			componentBasePrices[i] = catalog.getProductBasePrice(i);
		}
		int priceSum;
		for (int i = 0; i < numComputers; i++){
			priceSum = 0;
			int[] comps = bom.getComponents(i);
			for (int j = 0; j < 4; j++){
				componentIndexList[i][j] = catalog.getIndexFor(comps[j]);
				priceSum += catalog.getProductBasePrice(componentIndexList[i][j]);
			}
			computerBasePrices[i] = priceSum;
			todaysPrices[i] = new Vector();
		}
	}

	//store the RFQs
	private void handleCRFQ(int sender, int receiver, RFQBundle content) {
		for (int i = 0, n = content.size(); i < n; i++) {
			RFQInfo rfq = new RFQInfo();
			rfq.dueDate = content.getDueDate(i);
			rfq.reservePrice = content.getReservePricePerUnit(i);
			rfq.penalty = content.getPenalty(i);
			rfq.productID = content.getProductID(i);
			rfq.quantity = content.getQuantity(i);
			rfq.rfqID = content.getRFQID(i);
			rfq.name = participants[sender].getName();
			customerRFQs.put(new Integer(rfq.rfqID), rfq);
		}
		
	}
	
	/**
	 * Stores offers from customers.
	 */
	private void handleCOffer(int sender, int receiver, OfferBundle content){ 

		for (int i = 0, n = content.size(); i < n; i++) {
			RFQInfo rfq = (RFQInfo) customerRFQs.get(new Integer(content.getRFQID(i)));
			OfferInfo offer = new OfferInfo(content.getOfferID(i), content.getQuantity(i),
					content.getDueDate(i), (int)(content.getUnitPrice(i)), rfq, getComputerIndex(rfq.productID));
	
			customerOffers.put(sender + "-" + offer.offerID, offer); 							
		}
	}

	/**
	 * Stores orders from customers.
	 */
	private void handleCOrder(int sender, int receiver, OrderBundle content) {//order sent from customer
		for (int i = 0, n = content.size(); i < n; i++) {
			int orderID = content.getOrderID(i);
			OfferInfo offer = (OfferInfo) customerOffers.get(receiver + "-" + content.getOfferID(i));
			OrderInfo order = new OrderInfo(offer, orderID);
			customerOrders[date-1].add(order);
			todaysPrices[offer.index].add(new Integer(order.price));
		}
	}

	/**
	 * Stores RFQs to suppliers.
	 */
	private void handleSRFQ(int sender, int receiver, RFQBundle content) {//RFQ to supplier
		for (int i = 0, n = content.size(); i < n; i++) {
			RFQInfo rfq = new RFQInfo();
			rfq.dueDate = content.getDueDate(i);
			rfq.reservePrice = content.getReservePricePerUnit(i);
			rfq.penalty = content.getPenalty(i);
			rfq.productID = content.getProductID(i);
			rfq.quantity = content.getQuantity(i);
			rfq.rfqID = content.getRFQID(i);
			rfq.name = participants[sender].getName();
			supplierRFQs.put(receiver + "-" + rfq.rfqID, rfq);
			
			//check if it is an rfq for which a future prediction is needed (it will have a negative ID)
			//if so, put it in its own bundle and store
			if (rfq.rfqID < 0){
				RFQBundle b = new RFQBundle();
				b.addRFQ(rfq.rfqID, rfq.productID, rfq.quantity, rfq.reservePrice, rfq.dueDate, rfq.penalty);
				Message m = new Message (participants[sender].getName(), participants[receiver].getName(), b);
				futureRFQs[date].add(m);
			}
		}
	}
	
	/**
	 * Stores offers from suppliers.
	 */
	private void handleSOffer(int sender, int receiver, OfferBundle content){

		for (int i = 0, n = content.size(); i < n; i++) {
			RFQInfo rfq = (RFQInfo) supplierRFQs.get(sender + "-" + content.getRFQID(i));
			OfferInfo offer = new OfferInfo(content.getOfferID(i), content.getQuantity(i),
										content.getDueDate(i), (int)(content.getUnitPrice(i)), rfq, getComponentIndex(rfq.productID));
			supplierOffers[date-1].add(offer); 							
		}
	}

	/**
	 * Calculates the median prices for each computer type each day.
	 *
	 */
	private void getMedianPrices(){
		Integer[] a;
		for (int i = 0; i < 16; i++){
			if (todaysPrices[i].size() > 0){
				a = new Integer[todaysPrices[i].size()];
				todaysPrices[i].toArray(a);
				Arrays.sort(a);
				if (todaysPrices[i].size() % 2 == 1)
					medianPrice[i][date-1] = a[(a.length-1)/2].intValue() / (double) computerBasePrices[i];
				else //average 2 middle values
					medianPrice[i][date-1] = (a[a.length/2 - 1].intValue() + 
							a[a.length/2].intValue()) / 2 / (double) computerBasePrices[i];
				todaysPrices[i].clear();
			}
			else 
				medianPrice[i][date-1] = Double.NaN;
		}
	}
	
	
	/**
	 * Returns the index for the productID.
	 */
	private int getComponentIndex(int productID) {
		for (int i = 0; i < 10; i++) {
			if (componentIDs[i] == productID)
				return i;
		}
		return -1;
	}
  
	/**
	 * Returns the index for the productID.
	 */
	private int getComputerIndex(int productID) {
		for (int i = 0; i < 16; i++)
			if (computerIDs[i] == productID)
				return i;
		return -1;
	}

	/**
	 * Used for formatting output.
	 */
	private String pad(double d){
		String s = d + "     ";
		return s.substring(0, 5);
	}
}