//this includes much of the Main and LogManager log parsing classes from the original TAC SCM log tools, 
//which include the header below

/**
 * TAC Supply Chain Management Log Tools
 * http://www.sics.se/tac/		tac-dev@sics.se
 *
 * Copyright (c) 2001-2003 SICS AB. All rights reserved.
 *
 * SICS grants you the right to use, modify, and redistribute this
 * software for noncommercial purposes, on the conditions that you:
 * (1) retain the original headers, including the copyright notice and
 * this text, (2) clearly document the difference between any derived
 * software and the original, and (3) acknowledge your use of this
 * software in pertaining publications and reports.	SICS provides
 * this software "as is", without any warranty of any kind.	IN NO
 * EVENT SHALL SICS BE LIABLE FOR ANY DIRECT, SPECIAL OR INDIRECT,
 * PUNITIVE, INCIDENTAL OR CONSEQUENTIAL LOSSES OR DAMAGES ARISING OUT
 * OF THE USE OF THE SOFTWARE.
 *
 * -----------------------------------------------------------------
 *
 * Main
 *
 * Author	: Joakim Eriksson, Niclas Finne, Sverker Janson, Anders Sundman
 * Created : Tue Jul 15 13:16:15 2003
 * Updated : $Date: 2004/06/13 17:12:27 $
 *					 $Revision: 1.7 $
 */

import java.util.zip.GZIPInputStream;

import se.sics.tasim.logtool.LogReader;

import java.io.*;
import se.sics.isl.util.ArgumentManager;

import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

    
public class PredictionServerThread extends Thread{
	private ServerSocketHandler sh = null;
	private ArgumentManager config;
	private double[] results = new double[4]; //current computer, future computer, current component, future component RMS error
	private double[] resultSums = new double[4];
	private int gameCount = 0;
	private PrintStream output = null;
	private String participantName;
	private int startGame = 1;
	
	public PredictionServerThread(ServerSocketHandler sh, ArgumentManager config, String participantName) {
		super("PredictionServerThread");
		this.sh = sh;
		this.config = config;
		this.participantName = participantName;
	}

	public void run() {
		
		//open file for storing results
		DateFormat dateFormat = new SimpleDateFormat("MM-dd--HH-mm-ss");
		String date = dateFormat.format(new Date());
		try{
			output = new PrintStream(new FileOutputStream(participantName + "-" + date + ".txt"));
		} catch (Exception e){e.printStackTrace();}
		
		int startGame = sh.getStartGame();
		int currentGame = 0;
		
		//read the list of games
		boolean newGroup = true; //the next file represents a new group of agents
		BufferedReader f;
		String fileName = config.getProperty("gamelist");
		
		try{
					
			f = new BufferedReader(new FileReader(fileName));
			while (true){
				String s = getNextLine(f);
				if (s == null) //end of file
					break;
				else if (s.trim().equals("")){ //blank line 
					newGroup = true;
				}
				else{ //game file
					currentGame++;
					if (currentGame >= startGame){
						sh.writeAndRead("Starting game number " + currentGame + "\n");
						sh.writeAndRead(new NewGameMessage(newGroup));
						processDataFile(s);
						reportResults(s);
						sh.writeAndRead("Ended game number " + currentGame + "\n");
						newGroup = false;
					}
				}
			}
			f.close();
		} catch (Exception e) {
			e.printStackTrace();	
		}

		reportFinalResults();
		NewGameMessage n = new NewGameMessage(false);
		n.finished = true;
		sh.writeAndRead(n);
		sh.close();
		output.close();
	}
	
	/**
	 * Print results for a single game and update averages.
	 */
	void reportResults(String s){
		StringBuffer sb = new StringBuffer();
		sb.append("Results for " + s + " : " + "\n");
		sb.append("current computer prices RMS error: " + results[0] + "\n");
		sb.append("future computer prices RMS error: " + results[1] + "\n");
		sb.append("current component prices RMS error: " + results[2] + "\n");
		sb.append("future component prices RMS error: " + results[3] + "\n\n");
		output.println(sb);
		sh.writeAndRead(sb.toString());
		
		resultSums[0] += results[0];
		resultSums[1] += results[1];
		resultSums[2] += results[2];
		resultSums[3] += results[3];
		
		gameCount++;
	}
	
	/**
	 * Print results for all games.
	 */
	void reportFinalResults(){
		StringBuffer sb = new StringBuffer();
		sb.append("========================\nFINAL RESULTS\n========================\n");
		sb.append("current computer prices RMS error: " + resultSums[0]/gameCount + "\n");
		sb.append("future computer prices RMS error: " + resultSums[1]/gameCount + "\n");
		sb.append("current component prices RMS error: " + resultSums[2]/gameCount + "\n");
		sb.append("future component prices RMS error: " + resultSums[3]/gameCount + "\n\n");
		output.println(sb);
		sh.writeAndRead(sb.toString());
	}
	
	/**
	 * Read the next line from the file, ignoring lines starting with #
	 */
	static String getNextLine(BufferedReader f) throws Exception{
		String s;
		do {s = f.readLine();} while (s != null && (s.length() == 0 || s.charAt(0) == '#'));
		return s;
	}
	
	
	/**
	 * From the original LogManager class.  Processes a single file.
	 * Modified to call the parser directly.
	 */
	void processDataFile(String filename) {
		LogReader reader = null;
		try {
			reader = new LogReader(getDataStream(filename));
			output.println("Processing game " + reader.getSimulationID()
					+ " (" + filename + ')');
			try {
				String agentName = config.getProperty("agentName");
				boolean verbose = config.hasArgument("verbose");
				PredictionServerParser parser = new PredictionServerParser(reader, agentName, sh, output, verbose, results);
				parser.start();
			} finally {
				reader.close();
			}
		} catch (FileNotFoundException e) {
			System.err.println("could not find the game log file '" + filename + '\'');
		} catch (EOFException e) {
			if (reader == null || !reader.isCancelled()) {
				// Reader was not cancelled which means the file was truncated
				System.err.println("could not process the game log file '" + 
					filename + '\'');
				e.printStackTrace();
			}
		} catch (Exception e) {
			System.err.println("could not process the game log file '" + 
				filename + '\'');
			e.printStackTrace();
		}
	}
	
	/**
	 * From the original LogManager class.
	 */
	static InputStream getDataStream(String filename) throws IOException {
		File fp = new File(filename);
		if (!fp.exists()) {
			filename = filename.endsWith(".gz")
			? filename.substring(0, filename.length() - 3)
			: (filename + ".gz");
		}
		if (filename.endsWith(".gz")) {
			return new GZIPInputStream(new FileInputStream(filename));
		} else {
			return new FileInputStream(filename);
		}
	}
	
}





